#include "redef.h"

!>Computes or estimates the maximal timestep that can be used in explicit initial 
!!value computations 
!!
!!The estimate is also used to set dt for the performance optimization
!!\todo write a maxclimits routine which doesn't need full initialize_current_parall
module compute_dt
  use par_mod
  use discretization
  use coordinates
  use communications
  use phys_ini
  use aux_fields
  use eigen_parameters
  use RK_coefficients
  use calc_rhs, only: rhs_nl, rhs_f0, rhs_only_coll
  use external_contr, only: ExBrate, ExB_stime
  use numerical_damping
#ifdef WITHSLEPC
  use petsc_aux
  use slepc_aux
#endif

  implicit none
  public:: estimate_dt_max, set_dt_max, check_compute_dt
  public:: initialize_adapt_dt, adapt_dt, finalize_adapt_dt
#ifdef WITHSLEPC
  public:: dt_from_ev
#endif
  private

  !quantities needed for nonlinear time step adaption
  integer:: win_dts=100, win_est=25
  real,dimension(:),allocatable:: last_dts, last_estimates
  real:: ltav
  real::dt_vlasov=0.
  real,dimension(4)::dt_coll=0.
  !integer:: TIMEFILE=78

contains

  subroutine check_compute_dt
    if (comp_type=='IV') then
       if (calc_dt) then
          if(n_ev.eq.0) then
#ifdef oldparbc
             n_ev=1
#else
             n_ev=20
#endif
          end if
       else !calc_dt=.f.
          if (dt_max.le.0.) stop "specify dt_max or automatic calculation of time step"
       end if
    end if
  end subroutine check_compute_dt
  
  !>If calc_dt=.true., this routine either computes or estimates the maximum dt for 
  !>numerical stability of the time stepping scheme
  subroutine set_dt_max

    if (calc_dt) then   
#ifdef WITHSLEPC
       call compute_dt_max
#else
       call estimate_dt_max(.true.,dt_max)
#endif
    else 
       if ((mype.eq.0).and.(print_ini_msg)) then
          write(*,"(a,es10.3)") 'using time step limit from parameters: ',dt_max
       endif
    endif

  end subroutine set_dt_max

#ifdef WITHSLEPC
  !> Computes dt_max from the eigenvalues with maximal absolute value
  !!
  !!The eigenvalues with the largest absolute value are computed with SLEPc
  !!(the number can be changed with n_ev). Then, for each eigenvalue the maximal dt 
  !!necessary to keep this eigenvalue in the stability region of the time 
  !!stepping scheme is computed, the minimum of these values is then written to dt_max.
  !!The time stepping schemes supported at the moment are RK3, RK4, RK4M.
  !!Note that this method assumes that the eigenvalues are mostly aligned with the imaginary
  !!axis, so that the ones with the largest magnitude are actually the most critical for stability
  !!(see flm's PhD thesis). Note: if the spectrum is distorted, e.g. by large hyperdiffusion terms,
  !!the evs with the largest magnitude might not be the most critical for stability, in these cases 
  !!n_ev may have to be set considerably >1 (maybe 50 or so).
  !!\todo ky-parallelization (if any) is not exploited
  subroutine compute_dt_max

    double precision:: time_1, time_2
    complex,dimension(:),allocatable:: eigenvalues
    integer:: num_ev,num_ev_c
    integer:: ky0_ind_in, n_procs_y_in, n_procs_sim_in
    integer:: j, blocknum, ierr
    integer:: mpi_comm_world_save
    integer,dimension(:),allocatable:: kyind_for_ev
    logical:: sav_print_ini_msg
    real:: dt_max_in

    PERFON('comp_dt')

    in_timestep_comp = .true.

    sav_print_ini_msg=print_ini_msg

    if((mype.eq.0).and.print_ini_msg) then
       write(*,"(a)") '******************************************************'
       write(*,"(a)") 'Starting time step computation with PETSC/SLEPc'
       write(*,"(a)") '******************************************************'
    endif

    if ((.not.slepc_restartable()).and.(n_procs_y.gt.1).and.&
         &slepc_initialized_from_outside) then
       if (mype.eq.0) then
          write(*,"(a)") 'ERROR: the installed slepc version does not support'
          write(*,"(A)") '       repeated initialization as is required for '
          write(*,"(A)") '       n_procs_y > 1.'
          write(*,"(A)") '       Either install slepc > 3.1-p4 (recommended),'
          write(*,"(a)") '       use n_procs_y=1 or precompute dt_max.'
       endif
       stop
    endif
          

    call get_systime(time_1)
    !save present values
    ky0_ind_in=ky0_ind
    dt_max_in = dt_max
    n_procs_y_in=n_procs_y
    n_procs_sim_in=n_procs_sim

    !always use linear version of CalFullRhs for eigenvalue computation and neglect constant
    !f0 contribution
    rhs_nl=.false.
    rhs_f0=.false.
    rhs_only_coll=.false.
    num_ev=0

    !for y_local, the linear operator is block diagonal in the ky coordinate;
    !the time step limit for the whole operator is therefore the minimum of the
    !time step limits of the individual blocks (which can be computed much faster 
    !than the full system). 
    !Optimization and parallelization is not altered EXCEPT the n_procs_y,
    !which means that only 1/n_procs_y processors are used for the computation,
    !the rest idles
    if(y_local.and.(nky0.gt.1)) then
       n_procs_sim=n_procs_sim_in/n_procs_y_in
       n_procs_y=1
       nky0=1
       
       mpi_comm_world_save = MY_MPI_COMM_WORLD
       CALL set_my_mpi_comm_world(mpi_comm_world_save)

       if(mype.lt.n_procs_sim) then
          ! to further speed up the search for eigenvalues, we only check
          ! a maximum of 3 blocks. ky=ky0_ind, ky=ky0_ind+1 and ky=kymax.
          ! If nky0_in<=3, we check all ky modes given in the system
          
          if (nky0_in.gt.3) then
             blocknum=3
             allocate(kyind_for_ev(blocknum))
             do j=1,2
                kyind_for_ev(j)=ky0_ind_in+j-1
             enddo
             kyind_for_ev(3)=ky0_ind_in+nky0_in-1
          else
             blocknum=nky0_in
             allocate(kyind_for_ev(blocknum)) 
             do j=1,nky0_in
                kyind_for_ev(j)=ky0_ind_in+j-1
             enddo
          endif
          
          print_ini_msg = .false.
          call my_SlepcInitialize(MY_MPI_COMM_WORLD,.false.)

          do j=1,blocknum
             ky0_ind=kyind_for_ev(j)  
             call compute_relevant_evs(eigenvalues,num_ev)
          end do

          print_ini_msg = sav_print_ini_msg
          call dt_from_ev(eigenvalues,num_ev,.false.)
          
          if (allocated(eigenvalues)) deallocate(eigenvalues)

          if (arakawa_zv.and.hyp_on_h.and.hypz_opt) &
               call compute_dt_hypz(dt_max_in,dt_vlasov,blocknum,kyind_for_ev)

          if (coll_split) &
               call compute_dt_coll(dt_max_in,dt_vlasov,dt_coll,blocknum,kyind_for_ev,num_ev_c)

          if (slepc_restartable()) &
               call my_SlepcFinalize(.false.)
       endif

       !restore the values from the parameters file
       n_procs_sim=n_procs_sim_in

       call set_my_mpi_comm_world(mpi_comm_world_save)

       nky0=nky0_in
       ky0_ind=ky0_ind_in
       n_procs_y=n_procs_y_in

    else !y_local=.false. .or. nky0.eq.1

       if(mype.lt.n_procs_sim) then
          
          print_ini_msg = .false.
          call my_SlepcInitialize(MY_MPI_COMM_WORLD,.false.)
          call compute_relevant_evs(eigenvalues,num_ev)
          print_ini_msg = sav_print_ini_msg

          call dt_from_ev(eigenvalues,num_ev,.false.)

          if (allocated(eigenvalues)) deallocate(eigenvalues)

          blocknum=1
          allocate(kyind_for_ev(blocknum))
          kyind_for_ev=ky0_ind

          if (arakawa_zv.and.hyp_on_h.and.hypz_opt) &
               call compute_dt_hypz(dt_max_in,dt_vlasov,blocknum,kyind_for_ev)

          if (coll_split) &
               call compute_dt_coll(dt_max_in,dt_vlasov,dt_coll,blocknum,kyind_for_ev,num_ev_c)

          if (slepc_restartable()) &
               call my_SlepcFinalize(.false.)
       end if
    end if

    in_timestep_comp = .false.

    if (coll_split) then
       if ((mype.eq.0).and.print_ini_msg) write(*,"(a,i4,a,i4,a)") 'calculated ',num_ev,' + ',num_ev_c,' eigenvectors'
    else
       if ((mype.eq.0).and.print_ini_msg) write(*,"(a,i4,a)") 'calculated ',num_ev,' eigenvectors'
    endif 

    call mpi_bcast(dt_max,1,MPI_REAL_TYPE,0,MY_MPI_COMM_WORLD,ierr)
    if (arakawa_zv.and.hyp_on_h.and.hypz_opt) &
       &call mpi_bcast(hypz_compensation,1,MPI_LOGICAL,0,MY_MPI_COMM_WORLD,ierr)

    if (coll_split) then
       !the chosen RKCa coll_split_scheme (found on my_pey=0) is broadcasted
       call mpi_bcast(rkstages_coll,1,MPI_INTEGER,0,MY_MPI_COMM_WORLD,ierr)
       call mpi_bcast(dt_vlasov,1,MPI_REAL_TYPE,0,MY_MPI_COMM_WORLD,ierr)
       call mpi_bcast(dt_coll,size(dt_coll),MPI_REAL_TYPE,0,MY_MPI_COMM_WORLD,ierr)
       if (coll_split_scheme=='RKCa') then 
          print_ini_msg = .false.
          call switch_coll_scheme(rkstages_coll)
          print_ini_msg = sav_print_ini_msg
       endif
    endif

    call get_systime(time_2)
    time_ev=time_2-time_1
    
    
    if((mype.eq.0).and.print_ini_msg) then
       write(*,"(a,es12.3,a)") 'dt_max = ',dt_max ,' found with SLEPc'
       if (coll_split) write(*,"(a,es12.3,a,a,a)") '(dt_max = ',dt_vlasov ,' for vlasov part, ',timescheme,')'
       if (coll_split) write(*,"(a,es12.3,a,i1,a)")&
            &'(dt_max = ',dt_coll(rkstages_coll) ,' for collisions, RKC', rkstages_coll , '   )'
       write(*,"(a,f10.3,a)") 'time for computation of dt_max:',time_ev,' sec'
       write(*,"(a)") '******************************************************'
    endif

    call my_barrier()

    PERFOFF
  end subroutine compute_dt_max

  subroutine compute_dt_hypz(dt_max_in,dt_vlasov,blocknum,kyind_for_ev)
    implicit none
    complex,dimension(:),allocatable:: eigenvalues
    real,intent(in)::dt_max_in
    real,intent(inout)::dt_vlasov
    integer,intent(in)::blocknum
    integer,dimension(blocknum),intent(in)::kyind_for_ev
    integer:: num_ev
    logical :: sav_print_ini_msg
    integer::j

    if (allocated(eigenvalues)) deallocate(eigenvalues)
    dt_vlasov = dt_max
    dt_max = dt_max_in

    num_ev=0
    sav_print_ini_msg=print_ini_msg
    hypz_compensation=.false.
    print_ini_msg = .false.

    do j=1,blocknum
       ky0_ind=kyind_for_ev(j)          
       call compute_relevant_evs(eigenvalues,num_ev)
    end do
    print_ini_msg = sav_print_ini_msg
    if (mype==0.and.print_ini_msg) write(*,"(A)") 'dt_max without hyp_z compensation:'
    call dt_from_ev(eigenvalues,num_ev)

    if (dt_vlasov.gt.1.1*dt_max) then 
       hypz_compensation=.true.
       dt_max=dt_vlasov
       if (mype==0.and.print_ini_msg) write(*,"(A)") 'Keeping hyp_z compensation for larger timestep.'
    else
       dt_vlasov=dt_max
    endif
    if (allocated(eigenvalues)) deallocate(eigenvalues)

  end subroutine compute_dt_hypz



  subroutine compute_dt_coll(dt_max_in,dt_vlasov,dt_coll,blocknum,kyind_for_ev,num_ev)
    implicit none
    complex,dimension(:),allocatable:: eigenvalues
    real,intent(in)::dt_max_in
    real,intent(inout)::dt_vlasov
    real,dimension(max_rkstages_coll),intent(inout)::dt_coll
    integer,intent(in)::blocknum
    integer,dimension(blocknum),intent(in)::kyind_for_ev
    integer,intent(out)::num_ev
    logical :: sav_print_ini_msg
    integer::s

    if (allocated(eigenvalues)) deallocate(eigenvalues)
    dt_vlasov = dt_max
    dt_max = dt_max_in

    num_ev=0
    sav_print_ini_msg=print_ini_msg
    rhs_only_coll = .true.
    print_ini_msg = .false.
    !take maximum ky0_ind (assuming that this is relevant, if spatial diffusion is on)
    ky0_ind=kyind_for_ev(blocknum)
    call compute_relevant_evs(eigenvalues,num_ev)
    print_ini_msg = sav_print_ini_msg
    rhs_only_coll = .false.

    if (coll_split_scheme=='RKCa') then
       !theoretical limits for eps=0.05 damped shifted Chebychev Polynomials
       dt_coll(2) = 3.90419518*dt_coll(1)
       dt_coll(3) = 8.74400189*dt_coll(1)
       dt_coll(4) = 15.6167807*dt_coll(1)

       !compute the dt_coll array
       do rkstages_coll=1,max_rkstages_coll
          call dt_from_ev(eigenvalues,num_ev,.true.)
          dt_coll(rkstages_coll)=dt_max
       enddo

       !find appropriate rkstages_coll (default:1)
       !goal is to minimize stages and obtain a larger timestep than dt_vlasov/1.2
       !this 1.2 tries to account for the extra numerical effort
       rkstages_coll=1
       if (dt_coll(1).gt.0.and.dt_vlasov.gt.0) then
          rkstages_coll=max_rkstages_coll
          do s=max_rkstages_coll,1,-1
             if ((dt_coll(s).gt.dt_vlasov/1.2).and.(dt_coll(s).gt.0))&
                  &rkstages_coll=s
          enddo
       endif
       if ((mype.eq.0).and.print_ini_msg) write(*,"(a,i1)")&
            &'adaptive coll_split_scheme: we chose RKC',rkstages_coll
    else
       call dt_from_ev(eigenvalues,num_ev,.true.)
       dt_coll = dt_max
    endif

    if (dt_vlasov.gt.0.and.dt_coll(rkstages_coll).gt.0) dt_max = min(dt_vlasov,dt_coll(rkstages_coll))
    if (dt_vlasov.lt.0.and.dt_coll(rkstages_coll).gt.0) dt_max = dt_coll(rkstages_coll)
    if (dt_vlasov.gt.0.and.dt_coll(rkstages_coll).lt.0) dt_max = dt_vlasov

    if (allocated(eigenvalues)) deallocate(eigenvalues)

  end subroutine compute_dt_coll

  subroutine compute_relevant_evs(eigenvalues,num_ev)
    implicit none
    complex,dimension(:),allocatable:: eigenvalues
    integer:: num_ev
    integer:: iev, new_evs
    complex,dimension(:),allocatable:: old_eigenvalues
    call initialize_current_parall
    call initialize_calc_aux_fields
    
    call initialize_petsc_mat(.false.)
    call initialize_slepc_eps(.false.)
    call EPSSetFromOptions(eps_obj,globerr)
    PERFON('evsolver')
    call EPSSolve(eps_obj,globerr)
    PERFOFF
    
    call EPSGetConverged(eps_obj,new_evs,globerr)

    if (new_evs.eq.0) then
       if (mype.eq.0) write(*,"(a,G12.4,a)") '***no eigenvalues computed for ky=',ky,'!!***'
       call EPSGetIterationNumber(eps_obj,it_ev,globerr)
       if (mype.eq.0) write(*,"(a,i6)") 'number of iterations:',it_ev
    else
       if(num_ev.eq.0) then
          allocate(eigenvalues(0:new_evs-1))
       else
          allocate(old_eigenvalues(0:num_ev-1))
          old_eigenvalues=eigenvalues
          deallocate(eigenvalues)
          allocate(eigenvalues(0:num_ev+new_evs-1))
          eigenvalues(0:num_ev-1)=old_eigenvalues
          deallocate(old_eigenvalues)
       endif
       do iev=0,new_evs-1
          call EPSGetEigenpair(eps_obj,iev,eigenvalues(iev+num_ev),&
               &PETSC_NULL_OBJECT,glob_g,PETSC_NULL_OBJECT,globerr)
       enddo
       num_ev=num_ev+new_evs
    endif
    
    call finalize_slepc_eps
    call finalize_petsc_mat
             
    call finalize_current_parall
    call finalize_calc_aux_fields

  end subroutine compute_relevant_evs


  !>Calculates the dt_max for all eigenvalues in ev for the chosen time stepping
  !!scheme writes the minimum to the global variable dt_max
  subroutine dt_from_ev(ev,num_ev,for_coll_scheme)
    implicit none
    integer,intent(in):: num_ev !<number of eigenvectors that is passed
    complex,dimension(0:num_ev-1),intent(in):: ev !<array of the complex eigenvalues
    logical,intent(in),optional::for_coll_scheme
    logical::for_coll_scheme_
    real:: dt_final, dt_ev
    integer:: nr

    if (present(for_coll_scheme))then
       for_coll_scheme_=for_coll_scheme
    else
       for_coll_scheme_=.false.
    endif

    !set initial value
    dt_final=100.

    !compute dt_max for each eigenvalue and find the smallest
    do nr=0,num_ev-1
       call compute_dt_from_ev(ev(nr),dt_ev,for_coll_scheme_)
       if (dt_ev.ne.-1.) dt_final=min(dt_final,dt_ev)
    enddo

    !round to 3 significant digits
    dt_final=round_dt(0.99*dt_final)
    if (dt_final.ne.99.) then 
       dt_max=dt_final
       if ((mype.eq.0).and.(print_ini_msg)) then
       !print the result: did we find a reasonable value for dt_max?
          if (for_coll_scheme_) then
             write(*,"(a,es10.3,a,i1)") &
                & 'set time step limit to ',dt_final,' for collisions, RKC',rkstages_coll
          else
             write(*,"(a,es10.3,a,a)") 'set time step limit to ',dt_final, ' for timescheme ', timescheme
          endif
          if (which_ev.ne.'none') write(*,"(a)") '**********************************'
       endif
    else
        if (for_coll_scheme_) then
           if (mype.eq.0) write(*,"(A,a,a)") &
              & 'problem in time step calculation for collisions, RKC',rkstages_coll,' : dt_max not adapted!!'
        else
           if (mype.eq.0) write(*,"(A,a,a)") &
                & 'problem in time step calculation for timescheme ',timescheme,': dt_max not adapted!!'
        endif
    endif

  end subroutine dt_from_ev
#endif  

  !>Rounds dt to 3 significant digits (always decreases, never increases 
  !!dt to maintain stability)
  real function round_dt(dt_ev)
    real:: dt_ev !<value to be rounded
    real:: dt_aux
    integer::digits

    digits = 0

    dt_aux=dt_ev
    do while (dt_aux.lt.100)
       dt_aux=dt_aux*10
       digits=digits+1
    end do
    round_dt=real(floor(dt_aux))/10**digits

  end function round_dt

  !!Estimates without actual computation

  !>Prints the estimates for the time step limits in the various directions used
  !!by estimate_dt_max
  subroutine print_estimates
    logical:: write_pe
    real:: corrfac

    write_pe = ((mype==0).and.(print_ini_msg))

    if (write_pe) then
       write(*,*)
       write(*,"(A)") &
            &"=========== estimates for time step limit ==========="
       select case (parscheme)
       case('c4th')
          corrfac=1.4
       case('c6th')
          corrfac=1.6
       case default
          corrfac=1.0
       end select
       write(*,"(2(a,es10.3))") "CFL for parallel dynamics: dt<",&
            & dz/corrfac/maxclimit(3)*rk_corr(1)
       if (maxclimit(4).gt.0.0) then
          write(*,"(A,es10.3)")  "CFL for parallel velocity: dt<",&
               dv/1.4/maxclimit(4)*rk_corr(1)
       endif
       write(*,*)
       
       if ((maxclimit(1)>0).and.(maxval(kx).gt.0)) &
            & write(*,"(a,es10.3)") "linear x advection limit:  dt<",&
            & rk_corr(1)/(maxclimit(1)*maxval(kx))

       if ((maxclimit(2)>0).and.(kjmax.ne.0)) &
            & write(*,"(a,es10.3)") "linear y advection limit:  dt<",&
            & rk_corr(1)/(maxclimit(2)*kjmax)

       if(hyp_x.gt.0.) then
          if (yx_order) then
             write(*,"(a,es10.3)") "radial diffusion limit:    dt<",&
                  &rk_corr(2)/(hyp_x*djdiff_max)
          else
             write(*,"(a,es10.3)") "radial diffusion limit:    dt<",&
                  &rk_corr(2)/(hyp_x*didiff_max)
          endif
       endif
       
       if(hyp_y.gt.0.) then
          if (yx_order) then
             write(*,"(a,es10.3)") "binormal diffusion limit:  dt<",&
                  &rk_corr(2)/(hyp_y*didiff_max)
          else
             write(*,"(a,es10.3)") "binormal diffusion limit:  dt<",&
                  &rk_corr(2)/(hyp_y*djdiff_max)
          endif
       endif
       
       if(hyp_perp.gt.0.) then
          write(*,"(a,es10.3)") "perpend. diffusion limit:  dt<", &
               &rk_corr(2)/(hyp_perp*dperpdiff_max)
       endif
       
       if ((hyp_z_order.eq.4).and.(hyp_z.gt.0.)) &
            & write(*,"(a,es10.3)") "parallel diffusion limit:  dt<",&
            &rk_corr(2)/abs(hyp_z)

       if (hyp_v.gt.0.) &
            & write(*,"(a,es10.3)") "v_par diffusion limit:     dt<",&
            &rk_corr(2)/abs(hyp_v)

    endif
 
  end subroutine print_estimates

  !>Rough estimate of the maximum time step, the different advection terms are analyzed independently
  !!and the minimum is returned
  !!
  !!maxclimit must have been called befors so that maxclimitx/maxclimity are set
  !!\todo Routine doesn't take into account the stability regions of the different timeschemes
  !!\todo Adapt to nonlocal code
  subroutine estimate_dt_max(withoutput,edt)
    logical,intent(in):: withoutput !<switch whether estimates are printed to standard output
    real,intent(out):: edt !<estimate
    real::    cour_min=100.
    logical:: finalize_on_exit
    real:: vxmax, vymax, vzmax, vvmax

    !initialize RK scheme to get rk_corr
    call initialize_RK_coefficients

    !initialize maxclimitx/maxclimity if it hasn't been done already in perf_opt
    finalize_on_exit=.false.
    if (.not.allocated(kx)) then
       call initialize_current_parall
       finalize_on_exit=.true.
    endif


    !compute the advection limits in the different directions
    vxmax=0.
    vymax=0.
    vzmax=0.
    vvmax=0.
    
    !x
    IF (maxclimit(1).gt.0) then
       if (x_local) then
          if (maxval(kx).gt.0) &
               vxmax=(maxclimit(1)*maxval(kx))
       else
          !ADAPT THIS FACTOR FOR DERIVATIVES DIFFERENT FROM C4TH!
          vxmax = 1.4/deli*maxclimit(1)
       endif
    endif

    !y
    if ((maxclimit(2).gt.0).and.(kjmax.ne.0)) then
       vymax = maxclimit(2)*kjmax
    endif

    !z 
    !additional factor 1.4/1.6 for 4th/6th order finite differencing schemes
    !has been found experimentally but should be somewhere in the literature/
    !can probably derived after some thinking..
    select case (parscheme)
    case('c4th')
       vzmax=1.4/dz*maxclimit(3)
    case('c6th')
       vzmax=1.6/dz*maxclimit(3)
    case default
       vzmax=1./dz*maxclimit(3)
    end select
    
    !v_parallel
    if (maxclimit(4).gt.0.0) vvmax=1.4/dv*maxclimit(4)
    
    !in case vmax is zero, estimate dt ~ 1e-4
    if ((vxmax+vymax+vzmax+vvmax) .le. epsilon(vxmax)) vxmax=10000.0

    cour_min = rk_corr(1)/(vxmax+vymax+vzmax+vvmax)

    !diffusive limits
    if(hyp_x.gt.0.) then
       if (yx_order) then
          cour_min = min(cour_min, rk_corr(2)/(hyp_x*djdiff_max))
       else
          cour_min = min(cour_min, rk_corr(2)/(hyp_x*didiff_max))
       endif
    endif

    if(hyp_y.gt.0.) then
       if (yx_order) then
          cour_min = min(cour_min, rk_corr(2)/(hyp_y*didiff_max))
       else
          cour_min = min(cour_min, rk_corr(2)/(hyp_y*djdiff_max))
       endif
    endif
    
    if(hyp_perp.gt.0.) then
       cour_min = min(cour_min, rk_corr(2)/(hyp_perp*dperpdiff_max))
    endif     

    if ((hyp_z_order.eq.4).and.(hyp_z.gt.0.)) then
       cour_min = min(cour_min, rk_corr(2)/abs(hyp_z))
    endif

    if (hyp_v.gt.0.) then
       cour_min = min(cour_min, rk_corr(2)/abs(hyp_v))
    endif

    Call my_real_min_to_all(cour_min)

    edt=cour_min

    if (withoutput) then
       call print_estimates      
       if (mype.eq.0) then
          write(*,"(a,es10.3)") 'set time step limit to ',edt
       endif
    endif

    if(finalize_on_exit) call finalize_current_parall

  end subroutine estimate_dt_max

  subroutine initialize_adapt_dt
    allocate(last_dts(0:win_dts-1),last_estimates(0:win_est-1))
    last_dts=dt
    last_estimates=dt
    ltav=0.
    !if(mype.eq.0) open(TIMEFILE, file=trim(diagdir)//'/timeest', form='formatted',&
    !        status='replace', position='rewind')
  end subroutine initialize_adapt_dt

  !!\todo the definition of nl_dt should be generalized for the global case
  subroutine adapt_dt
    real:: nl_dt, dt_inst, dt_est, dt_all, dt_GyroLES
    real, dimension(win_est):: sortdts
    real, dimension(win_est-2):: deri
    real:: av, var, normvar

    !linear part is neglected (and compensated by the courant factor),
    !if it is added here the time step is way too small
    if (parallel_nl) then
       nl_dt=rk_corr(1)/(ve_max(1)*kx(hkx)+ve_max(2)*kjmax+ve_pnl_max/dv)
    else
       nl_dt=rk_corr(1)/(ve_max(1)*kx(hkx)+ve_max(2)*kjmax)
    endif

    last_estimates(mod(itime-1,win_est)) = nl_dt

    !some statistical analysis to detect instabilities
    sortdts=cshift(last_estimates,mod(itime-1,win_est)-win_est+1)
    !2nd derivative
    deri=2*sortdts(2:win_est-1)-sortdts(3:win_est)-sortdts(1:win_est-2)
   
    !variance of the sample
    av=sum(deri)/(win_est-2)
    var=sum((deri-av)**2)/(win_est-3)

    !normalize variance to the old average
    if (((itime.le.win_est+1).or.(ltav.lt.(1000*epsilon(ltav)))).or. &
         &((abs(ExBrate).gt.epsilon(ExBrate)).and.(time.ge.ExB_stime))) then
       normvar=0. 
    else
       normvar=var/ltav
    end if

    !limit the reduction factor of the variance term to 0.3
    if (normvar.gt.50) normvar=50

    !update long term average
    if (itime.gt.win_est) then
       if(itime.gt.win_est+500) then
          ltav=0.998*ltav+0.002*var
       else
          ltav=((itime-win_est-1)*ltav+var)/(itime-win_est)
       end if
    end if
    
    if (courant==0.0) then
       !variance analysis: reduce dt when nl_dt decreases quickly (likely an instability),
       dt_est = nl_dt/cosh(normvar/4)**0.1
       !add linear eigenvalues and max. ve shift (worst case scenario)
       dt_all = 1.0/(1.0/dt_max+1.0/dt_est)
    else
       !completely neglect linear part and compensate by courant factor
       dt_est = courant*nl_dt
       !combine linear, nonlinear and variance constraint
       dt_all=min(dt_est/cosh(normvar/4)**0.1,dt_max)
    endif

    if (coll_split_scheme=='RKCa') call adapt_coll_scheme(dt_all,dt_est,normvar)

    last_dts(mod(itime-1,win_dts))=dt_all

    dt_inst=minval(last_dts)

    if (dt_inst/dt.lt.0.8) then
       dt=dt_inst
    elseif(dt_inst/dt.gt.1.2) then
       !ramp up dt only gradually
       dt=1.2*dt
       !modify last_dts to store this value for win_dts timesteps
       last_dts(mod(itime-1,win_dts))=1.2*dt
    endif

    !Modification to GyroLES 
    if (GyroLES) then
        dt_GyroLES = min(rk_corr(2)/(maxval(hyp_x_spec)*didiff_max),rk_corr(2)/(maxval(hyp_y_spec)*djdiff_max))
        dt=min(dt,dt_max,dt_GyroLES)
    else
        !if(mype.eq.0) Write(TIMEFILE, "(5ES20.8)") time,nl_dt,normvar,dt_all,ltav
        dt=min(dt,dt_max) ! required in some cases to bridge the gap
                          ! between 0.8 and 1.0 if dt_inst=dt_max<dt
    endif

  end subroutine adapt_dt


  subroutine adapt_coll_scheme(dt_all,dt_est,normvar)
    implicit none
    real,intent(in)::dt_all,dt_est,normvar
    real::dt_s  !timestep for different number of stages
    real::dt_max_new
   
    !dt_vlasov has to be determined with SLEPC, otherwise we will not adapt dt
    if (dt_vlasov.le.0) return

    !if current nonlinear dt is significantly (by 1.21) smaller than the maximum
    !allowed timestep of a lower stages coll_split_scheme,
    !we reduce the number of stages
    !1.21 is slightly larger than 1.2, the threshold for linear scheme adaption
    if (rkstages_coll.gt.1) then
       if (dt_coll(rkstages_coll-1).gt.(1.21*dt_all)) then
          rkstages_coll=rkstages_coll-1
          call switch_coll_scheme(rkstages_coll)
          dt_max=min(dt_vlasov,dt_coll(rkstages_coll))
          if (mype.eq.0) write(*,"(a,i1,a,es10.3)") 'adapting coll_split_scheme to ', rkstages_coll,' stages at time=',time
          if (mype.eq.0) write(*,"(a,es10.3)") '   reset dt_max to ', dt_max
       endif
    endif

    !if nonlinear dt_all is significantly (1.21 times) larger with a higher stages coll_split_scheme,
    !we increase the number of stages (this should be rare..)
    !1.21 is slightly larger than 1.2, the threshold for linear scheme adaption
    if (rkstages_coll.lt.max_rkstages_coll) then
       !linear dt for higher scheme
       dt_max_new=min(dt_vlasov,dt_coll(rkstages_coll+1))

       !combine nonlinear, higher scheme linear and variance constraints
       if(courant==0) then
          dt_s=1.0/(1.0/dt_max_new + 1.0/dt_est)
       else
          dt_s=min(dt_est/cosh(normvar/4)**0.1,dt_max_new)
       endif

       if (dt_s.gt.1.21*dt_all) then
          rkstages_coll=rkstages_coll+1
          call switch_coll_scheme(rkstages_coll)
          dt_max=dt_max_new
          !dt_all is not adapted (this is left for next timestep)
          if (mype.eq.0) write(*,"(a,i1,a,es10.3)") 'adapting coll_split_scheme to ', rkstages_coll,' stages at time =',time
          if (mype.eq.0) write(*,"(a,es10.3)") '   reset dt_max to ', dt_max
          if (mype.eq.0) write(*,"(a,es10.3,a,es10.3)") '   nonlinear estimate increases from dt=', dt_all, ' to dt=',dt_s 
       endif
    endif

  end subroutine adapt_coll_scheme

  subroutine finalize_adapt_dt
    !if(mype.eq.0) close(TIMEFILE)
    deallocate(last_dts,last_estimates)
  end subroutine finalize_adapt_dt


end module compute_dt
