#include "redef.h"
#include "intrinsic_sizes.h"
!>Contains routines to diagnose the eigenvalue spectrum in a nonlinear run
!!(linearizes the right hand side by taking a snapshot of the electromagnetic fields)
!/todo use this module in diag.F90 (remove circular dependency with slepc.aux)
Module diag_nl_eigenvalues
  Use aux_fields
  use calc_rhs, only: rhs_nl, rhs_f0, rhs_only_coll, rhs_nlev
  Use communications
  use discretization
  use eigen_parameters
  !use coordinates
  !use RK_coefficients
  Use file_io
  Use par_mod
  !use phys_ini
#ifdef WITHSLEPC
  use compute_dt, only: dt_from_ev
  use petsc_aux
  use slepc_aux
#endif

  Implicit None

  !The basic diag interface
  PUBLIC :: initialize_all_diags_nlev, exec_all_diags_nlev, &
       &finalize_all_diags_nlev,mem_est_all_diags_nlev, &
       &check_diag_nlev,&
       &istep_nlev, nlev_stime, n_nlev

  PRIVATE 

  integer:: NLEVFILE
  integer:: istep_nlev=-1,n_nlev = 0
  real:: nlev_stime = -1
  Character(Len=8):: filestat='replace', filepos='rewind'

Contains

  !> check input parameters for this module
  Subroutine check_diag_nlev

#ifndef WITHSLEPC
    if (istep_nlev.gt.0) then
       if (mype==0) write(*,'(a)') 'no SLEPC/PETSc linked: we set istep_nlev = 0'
       istep_nlev=0
    endif
#endif
    if ((istep_nlev.gt.0).and.(.not.nonlinear)) then
       if (mype==0) Write(*,'(A)') 'switching off diag_nlev (only possible for nonlinear runs)'
       istep_nlev=0
    endif
    
    if ((istep_nlev.gt.0).and.(.not.xy_local)) then
       if (mype==0) Write(*,'(A)') 'switching off diag_nlev (currently not implemented for global runs)'
       istep_nlev=0
    endif

    if (istep_nlev.gt.0) then
       !setting the number of nonlinear eigenvalues
       if(n_nlev.le.0) then
          if (n_ev.gt.0) then 
             n_nlev = n_ev
          else
             n_nlev = 20
          endif
          if (mype==0) Write(*,'(a,i6,a)') 'using ',n_nlev,' eigenvalues for diag_nlev'
       end if
    end if

  End Subroutine check_diag_nlev


  !>Give an estimate of the memory requirements of this module
  !!(no big memory demand at the moment)
  Real Function mem_est_all_diags_nlev(mem_req_in)
    real:: mem_req_in
    real:: mem_loc=0

    !if (istep_nlev.gt.0) mem_loc = mem_loc+0.0

    mem_est_all_diags_nlev=mem_req_in+mem_loc

  End Function mem_est_all_diags_nlev


  Subroutine initialize_all_diags_nlev

    if (istep_nlev.gt.0) call initialize_diag_nlev

  End Subroutine initialize_all_diags_nlev

  Subroutine exec_all_diags_nlev(itime,time,first)
    integer, intent(in) :: itime
    real, intent(in) :: time
    logical, intent(in) :: first
    logical::rhs_nl_in
    
#ifdef WITHSLEPC
    if (istep_nlev.gt.0) then
       if(first) then
          !one set of eigenvalues for the linear case
          rhs_nl_in =  rhs_nl
          rhs_nl =  .false.
          call diag_nlev(itime,time,first)
          rhs_nl=rhs_nl_in
       else
          if ((modulo(itime,istep_nlev) == 0).and.(time.ge.nlev_stime)) &
               &call diag_nlev(itime,time,first)
       endif
    endif
#endif

  End Subroutine exec_all_diags_nlev


  Subroutine finalize_all_diags_nlev

    if (istep_nlev.gt.0) call finalize_diag_nlev

  End Subroutine finalize_all_diags_nlev

!*****************************************************************

  Subroutine initialize_diag_nlev
    Implicit None
    if (write_std) then
       if(mype.eq.0) open(NLEVFILE, file=trim(diagdir)//'/eigenvalues_nl'//&
            &''//trim(file_extension),&
            form='formatted', status=filestat, position=filepos)  
    endif

  End Subroutine initialize_diag_nlev

#ifdef WITHSLEPC
  Subroutine write_evfile(eigenvalues,num_ev,dt_nl,time,first)
    implicit none
    integer,intent(in)::num_ev
    complex,dimension(0:num_ev-1),intent(in)::eigenvalues
    real,intent(in)::dt_nl
    real,intent(in)::time
    logical,intent(in)::first
    integer::i

    !output to file
    if(write_std) then
       if (mype.eq.0) then
          if(first)then
             write(NLEVFILE, "(a,ES12.3,a)") '#linear eigenvalues (re/im part),  (dt_max = ',dt_max,')'
          else
             write(NLEVFILE, "(a,f10.3,a,ES10.3,a,ES10.3,a)") &
                &'#nonlinear eigenvalues (re/im part) at time ', time, ' (dt_nl = ',dt_nl,', dt = ',dt,')'
          endif
          do i=0,num_ev-1
             write(NLEVFILE, "(8ES20.8)") eigenvalues(i)
          enddo
          write(NLEVFILE, "(a)") NEW_LINE('(a)')
       end if
       call flush(NLEVFILE)
    end if

  End Subroutine write_evfile
#endif
!*************************************************

  !>computes and writes nonlinear eigenvalues
  !!this is basically the same as compute_dt, but using all ky modes
  !!Runge-Kutta timestep is computed(for comparison), but not adapted
  Subroutine diag_nlev(itime,time,first)
    implicit none
    integer,intent(in):: itime
    real,intent(in):: time
    logical,intent(in):: first
    double precision:: time_1, time_2
    complex,dimension(:),allocatable:: eigenvalues
    integer:: num_ev
!    integer:: ierr
!    integer:: mpi_comm_world_save
    logical:: sav_print_ini_msg, sav_rhs_f0
    real:: dt_max_in, dt_nl

#ifdef WITHSLEPC
    PERFON('diag_nl_eigenvalues')

    in_timestep_comp = .true.
    
    sav_print_ini_msg=print_ini_msg
    print_ini_msg = .false.
    !alway neglect constant f0 contribution to CalFullRhs for eigenvalue computation
    sav_rhs_f0 = rhs_f0
    rhs_f0=.false.

    if((mype.eq.0).and.print_ini_msg) then
       write(*,"(a)") '******************************************************'
       write(*,"(a)") 'Starting eigenvalue computation with PETSC/SLEPc'
       write(*,"(a)") '******************************************************'
    endif

    call get_systime(time_1)
    !save present values
    dt_max_in = dt_max

    num_ev=0

    call my_SlepcInitialize(MY_MPI_COMM_WORLD,.false.)
    call compute_nonlinear_evs(eigenvalues,num_ev)
    if (slepc_restartable()) call my_SlepcFinalize(.false.)

    if ((mype.eq.0).and.print_ini_msg) write(*,"(a,i4,a)") 'calculated ',num_ev,' eigenvectors'

    call dt_from_ev(eigenvalues,num_ev)

    !reset linear dt_max (also for first write_evfile)
    dt_nl = dt_max
    dt_max = dt_max_in

    call write_evfile(eigenvalues,num_ev,dt_nl,time,first)
    if (allocated(eigenvalues)) deallocate(eigenvalues)

    in_timestep_comp = .false.
    
    !call mpi_bcast(dt_max,1,MPI_REAL_TYPE,0,MY_MPI_COMM_WORLD,ierr)
    call get_systime(time_2)
    time_ev=time_2-time_1
    
    if((mype.eq.0).and.print_ini_msg) then
       write(*,"(a)") '******************************************************'
       write(*,"(a,es12.3,a)") 'dt_nl  = ',dt_nl ,' computed with SLEPc'
       write(*,"(a,es12.3,a)") 'dt     = ',dt ,' currently used in nonlinear run'
       write(*,"(a,es12.3,a)") 'dt_max = ',dt_max ,' for linear operator'
       write(*,"(a)") '******************************************************'
    endif
    print_ini_msg = sav_print_ini_msg
    rhs_f0 = sav_rhs_f0
    if (mype.eq.0) write(*,"(a,f10.3,a)") 'Time for computation of nonlinear eigenvalues:',time_ev,' sec'

    PERFOFF
#endif
  End Subroutine diag_nlev

!*************************************************
#ifdef WITHSLEPC
  subroutine compute_nonlinear_evs(eigenvalues,num_ev)
    implicit none
    complex,dimension(:),allocatable:: eigenvalues
    integer:: num_ev
    integer:: iev
    
    !use the linearized version of the nonlinearity in CalFulRhs
    rhs_nlev = .true.

    call initialize_petsc_mat(.false.,for_nlev=.true.)
    call initialize_slepc_eps(.false.,n_nlev)
    call EPSSetFromOptions(eps_obj,globerr)
    !PERFON('nl_evsol')
    call EPSSolve(eps_obj,globerr)
    !PERFOFF
    
    call EPSGetConverged(eps_obj,num_ev,globerr)

    if (num_ev.eq.0) then
       if (mype.eq.0) write(*,"(a)") '***no eigenvalues computed !!***'
       call EPSGetIterationNumber(eps_obj,it_ev,globerr)
       if (mype.eq.0) write(*,"(a,i6)") 'number of iterations:',it_ev
    else
       allocate(eigenvalues(0:num_ev-1))
       do iev=0,num_ev-1
          call EPSGetEigenpair(eps_obj,iev,eigenvalues(iev),&
               &PETSC_NULL_OBJECT,glob_g,PETSC_NULL_OBJECT,globerr)
       enddo
    endif
    
    call finalize_slepc_eps
    call finalize_petsc_mat(.false.)

    rhs_nlev = .false.

  end subroutine compute_nonlinear_evs
#endif

  Subroutine finalize_diag_nlev

    if (write_std.and.mype.eq.0) close(NLEVFILE)

  End Subroutine finalize_diag_nlev
  
!******************************************************************!!!

End Module diag_nl_eigenvalues
