#include "redef.h"
!>lagrange_interpolation contains subroutines to perform
!!a mid-point lagrange interpolation of order lag_order
!!and a filter to return to the orginal grid.
!!This library is used in the x-nonlocal version for
!!dealiasing in the nonlinearity and optionally
!!in the parallel boundary condition
MODULE lagrange_interpolation
  IMPLICIT NONE
  PUBLIC:: initialize_lag_coef, lag_interp, lag_filtering, finalize_lag_coef
  PUBLIC:: lag3interp, lag3deriv, lag3interp_2d
  PUBLIC:: lag3interp_complex
  PRIVATE

  INTERFACE lag3interp
     MODULE PROCEDURE lag3interp_scalar, lag3interp_array
  END INTERFACE

  INTERFACE lag3deriv
     MODULE PROCEDURE lag3deriv_scalar, lag3deriv_array
  END INTERFACE

  INTEGER:: li1,li2,li1da,li2da,lbida,ubida,lbida2,ubida2,lag_order
  REAL, DIMENSION(:),Allocatable:: Lag_coef

CONTAINS

  !> Initializes the Lagrange coefficients for mid-point
  !! lagrange interpolation using polynomials of order
  !! lag_order (i.e. using lag_order+1 grid points)
  !! \param a_li1 original array start index (x) without boundary points (b.p.)
  !! \param a_li2 original array end index (x) -"- 
  !! \param a_li1da start index (x) of interpolated array without b.p. (da for dealiasing)
  !! \param a_li2da end index (x) of -"-
  !! \param a_lbida start index (x) of original array with b.p. adapted for lag. interpol.
  !! \param a_ubida end index (s) of -"-
  !! \param a_lbida2 start index (x) of interpolated array with b.p. adapted for centered diff.
  !! \param a_ubida2 end index (x) of -"-
  subroutine initialize_lag_coef(a_li1,a_li2,a_li1da,a_li2da,&
         a_lbida, a_ubida, a_lbida2, a_ubida2, a_lag_order)
    INTEGER,INTENT(IN) :: a_li1,a_li2,a_li1da,a_li2da,&
         a_lbida, a_ubida, a_lbida2, a_ubida2, a_lag_order
    real, dimension(:), allocatable ::  xi_g
    integer :: Npi,i,j
    real :: xi_pos,dxi

    li1   = a_li1  ;   li2   = a_li2 ! first and last point, taken from main
    li1da = a_li1da;   li2da = a_li2da
    lbida = a_lbida;   ubida = a_ubida
    lbida2 = a_lbida2; ubida2 = a_ubida2

    lag_order = a_lag_order

    if (modulo(lag_order,2).eq.0) then
       stop 'lag_order : Only odd Lagrange interpolation order is allowed'
    end if

    ! Calculate the Lagrange coefficients
    ! 1. Setup a normalized subgrid which runs from 0 to 1 with
    ! Npi grid points
    Npi=lag_order+1
    allocate(xi_g(1:Npi),Lag_coef(1:Npi))
    !PRINT*,"Lag_coef allocated"
    dxi=1.0/lag_order
    do i=1,Npi
       xi_g(i)=(i-1)*dxi
    end do
    
    ! Compute Lagrange coefficients
    xi_pos=0.5     ! mid point interpolation
    Lag_coef=1
    Do i=1,Npi
       Do j=1,Npi
          if (j.ne.i) then 
             Lag_coef(i)=Lag_coef(i)*(xi_pos-xi_g(j))/(xi_g(i)-xi_g(j))
          end if
       end Do
    end Do
    deallocate(xi_g)
  end subroutine initialize_lag_coef
  

  !>Interpolates array fin and stores the result in fout
  !!
  !! \param fin input array with boundary points adapted for lag. interp.
  !! \param fout interpolated array
  !! \param l1 number of elements in remaining dimensions
  subroutine lag_interp(fin,fout,l1)
    integer, intent(in) :: l1
    COMPLEX, DIMENSION(lbida:ubida,0:*), INTENT(in) :: fin
    COMPLEX, DIMENSION(li1da:li2da,0:*), INTENT(out) :: fout
    INTEGER :: i, Npi,ind, l, base_ind
    
    PERFON('lag_int')

    Npi=lag_order+1
    do l=0,l1-1
       ! set all even elements of fout to the values of fin
       ind = li1da
       DO i=li1,li2
          ! even elements are copied
          fout(ind,l) = fin(i,l)

          ! odd elements are interpolated
          base_ind = i-Npi/2
          fout(ind+1,l) = SUM(Lag_coef*fin(base_ind+1:base_ind+Npi,l))
          
          ind = ind + 2
       END DO
    end do
    PERFOFF
    
  end subroutine lag_interp
  
  !> Filters interpolated input array, the result
  !! is stored in fout
  !!
  !! \param fin input array with boundary points adapted for finite diff. schemes
  !! \param fout filtered array
  !! \param l1  number of elements in remaining dimensions
  SUBROUTINE lag_filtering(fin,fout,l1,n_boundary_points)
    integer, intent(in) :: l1
    INTEGER, intent(IN) :: n_boundary_points
    complex, dimension(lbida2:ubida2,0:l1-1), intent(in) :: fin
    COMPLEX, DIMENSION(li1-n_boundary_points:li2+n_boundary_points,0:l1-1), INTENT(out) :: fout

    integer:: i, l, s, Npi, ind
    
    PERFON('lag_filter')

    Npi=lag_order+1
    do l=0,l1-1
       DO i=li1,li2
          fout(i,l)=0.5*fin(2*i-li1,l)
          do s=1,Npi
             ind=2*(i+s-Npi/2)-li1-1
             fout(i,l)=fout(i,l)+0.5*Lag_coef(s)*fin(ind,l)
          end do
       end do
    end do

    PERFOFF
  end subroutine lag_filtering



  !> Removes and deallocates all data structures used in this module
  subroutine finalize_lag_coef
    IF (ALLOCATED(Lag_coef))  THEN
       DEALLOCATE(Lag_coef)
    END IF
  end subroutine finalize_lag_coef

  !> Third order lagrange interpolation
  SUBROUTINE lag3interp_scalar(y_in,x_in,n_in,y_out,x_out)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: n_in
    REAL, DIMENSION(n_in), INTENT(IN) :: y_in,x_in
    REAL, INTENT(IN) :: x_out
    REAL, INTENT(OUT) :: y_out

    REAL, DIMENSION(1) :: xout_wrap, yout_wrap

    xout_wrap = x_out
    call lag3interp_array(y_in,x_in,n_in,yout_wrap,xout_wrap,1)
    y_out = yout_wrap(1)

  END SUBROUTINE lag3interp_scalar

  !> Third order lagrange interpolation
  subroutine lag3interp_array(y_in,x_in,n_in,y_out,x_out,n_out)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: n_in,n_out
    REAL, DIMENSION(n_in), INTENT(IN) :: y_in,x_in
    REAL, DIMENSION(n_out), INTENT(IN) :: x_out
    REAL, DIMENSION(n_out), INTENT(OUT) :: y_out

    REAL :: x,aintm,aint0,aint1,aint2,xm,x0,x1,x2
    INTEGER :: j,jm,j0,j1,j2
    INTEGER :: jstart,jfirst,jlast,jstep

    IF (x_in(n_in) > x_in(1)) THEN
       jstart=3
       jfirst=1
       jlast=n_out
       jstep=1
    ELSE
       jstart=n_in-2
       jfirst=n_out
       jlast=1
       jstep=-1
    END IF

    j1=jstart
    DO j=jfirst,jlast,jstep
       x=x_out(j)
       DO WHILE (x >= x_in(j1) .AND. j1 < n_in-1 .AND. j1 > 2) 
          j1=j1+jstep
       END DO

       j2=j1+jstep
       j0=j1-jstep
       jm=j1-2*jstep

       !...  extrapolate inside or outside

       x2=x_in(j2)
       x1=x_in(j1)
       x0=x_in(j0)
       xm=x_in(jm)

       aintm=(x-x0)*(x-x1)*(x-x2)/((xm-x0)*(xm-x1)*(xm-x2))
       aint0=(x-xm)*(x-x1)*(x-x2)/((x0-xm)*(x0-x1)*(x0-x2))
       aint1=(x-xm)*(x-x0)*(x-x2)/((x1-xm)*(x1-x0)*(x1-x2))
       aint2=(x-xm)*(x-x0)*(x-x1)/((x2-xm)*(x2-x0)*(x2-x1))

       y_out(j)=aintm*y_in(jm)+aint0*y_in(j0) &
            +aint1*y_in(j1)+aint2*y_in(j2)

    END DO

  END SUBROUTINE Lag3interp_array
  
  
  !> Third order lagrange interpolation for complex arrays
  SUBROUTINE lag3interp_complex(y_in,x_in,n_in,y_out,x_out,n_out)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: n_in,n_out
    COMPLEX, DIMENSION(n_in), INTENT(IN) :: y_in
    REAL, DIMENSION(n_in), INTENT(IN) :: x_in
    COMPLEX, DIMENSION(n_out), INTENT(OUT) :: y_out
    REAL, DIMENSION(n_out), INTENT(IN) :: x_out

    REAL :: x,aintm,aint0,aint1,aint2,xm,x0,x1,x2
    INTEGER :: j,jm,j0,j1,j2
    INTEGER :: jstart,jfirst,jlast,jstep

    IF (x_in(n_in) > x_in(1)) THEN
       jstart=3
       jfirst=1
       jlast=n_out
       jstep=1
    ELSE
       jstart=n_in-2
       jfirst=n_out
       jlast=1
       jstep=-1
    END IF

    j1=jstart
    DO j=jfirst,jlast,jstep
       x=x_out(j)
       DO WHILE (x >= x_in(j1) .AND. j1 < n_in-1 .AND. j1 > 2) 
          j1=j1+jstep
       END DO

       j2=j1+jstep
       j0=j1-jstep
       jm=j1-2*jstep

       !...  extrapolate inside or outside

       x2=x_in(j2)
       x1=x_in(j1)
       x0=x_in(j0)
       xm=x_in(jm)

       aintm=(x-x0)*(x-x1)*(x-x2)/((xm-x0)*(xm-x1)*(xm-x2))
       aint0=(x-xm)*(x-x1)*(x-x2)/((x0-xm)*(x0-x1)*(x0-x2))
       aint1=(x-xm)*(x-x0)*(x-x2)/((x1-xm)*(x1-x0)*(x1-x2))
       aint2=(x-xm)*(x-x0)*(x-x1)/((x2-xm)*(x2-x0)*(x2-x1))

       y_out(j)=aintm*y_in(jm)+aint0*y_in(j0) &
            +aint1*y_in(j1)+aint2*y_in(j2)

    END DO

  END SUBROUTINE lag3interp_complex


  !>2D interpolation 
  !\TODO check whether a "real" 2D interpolation would
  !! be more appropriate
  SUBROUTINE lag3interp_2d(y_in,x1_in,n1_in,x2_in,n2_in,&
       &y_out,x1_out,n1_out,x2_out,n2_out)

    IMPLICIT NONE
    
    INTEGER, INTENT(IN) :: n1_in,n2_in,n1_out,n2_out
    REAL, DIMENSION(n1_in,n2_in), INTENT(IN) :: y_in
    REAL, DIMENSION(n1_in) :: x1_in
    REAL, DIMENSION(n2_in) :: x2_in
    REAL, DIMENSION(n1_out), INTENT(IN) :: x1_out
    REAL, DIMENSION(n2_out), INTENT(IN) :: x2_out
    REAL, DIMENSION(n1_out,n2_out), INTENT(OUT) :: y_out    

    !local variables
    REAL, DIMENSION(n2_in) :: y2_in_tmp
    REAL, DIMENSION(n2_out) :: y2_out_tmp
    REAL, DIMENSION(n1_in,n2_out) :: y_tmp
    INTEGER :: i

    DO i=1,n1_in
       y2_in_tmp = y_in(i,:)
       call lag3interp(y2_in_tmp,x2_in,n2_in,&
            y2_out_tmp,x2_out,n2_out)
       y_tmp(i,:) = y2_out_tmp
    ENDDO
    
    DO i=1,n2_out
       call lag3interp(y_tmp(:,i),x1_in,n1_in,&
            y_out(:,i),x1_out,n1_out)
    END DO
    
  END SUBROUTINE lag3interp_2d

  !> Third order lagrange interpolation
  SUBROUTINE lag3deriv_scalar(y_in,x_in,n_in,dydx_out,x_out)

    IMPLICIT NONE

    INTEGER, INTENT(IN) :: n_in
    REAL, DIMENSION(n_in), INTENT(IN) :: y_in,x_in
    REAL, INTENT(IN) :: x_out
    REAL, INTENT(OUT) :: dydx_out

    REAL, DIMENSION(1) :: xout_wrap, dydxout_wrap

    xout_wrap = x_out
    call lag3deriv_array(y_in,x_in,n_in,dydxout_wrap,xout_wrap,1)
    dydx_out = dydxout_wrap(1)

  END SUBROUTINE lag3deriv_scalar



!>Returns Derivative based on a 3rd order lagrange interpolation
  subroutine lag3deriv_array(y_in,x_in,n_in,dydx_out,x_out,n_out)

    IMPLICIT NONE

    INTEGER :: n_in,n_out
    REAL, DIMENSION(n_in), INTENT(IN) :: y_in,x_in
    REAL, DIMENSION(n_out), INTENT(IN) :: x_out
    REAL, DIMENSION(n_out), INTENT(OUT) :: dydx_out

    REAL :: x,aintm,aint0,aint1,aint2,xm,x0,x1,x2
    INTEGER :: j,jm,j0,j1,j2
    INTEGER :: jstart,jfirst,jlast,jstep

    IF (x_in(n_in) > x_in(1)) THEN
       jstart=3
       jfirst=1
       jlast=n_out
       jstep=1
    ELSE
       jstart=n_in-2
       jfirst=n_out
       jlast=1
       jstep=-1
    END IF
    
    j1=jstart
    DO j=jfirst,jlast,jstep
       x=x_out(j)
       DO WHILE (x >= x_in(j1) .AND. j1 < n_in-1 .AND. j1 > 2) 
          j1=j1+jstep
       END DO

       j2=j1+jstep
       j0=j1-jstep
       jm=j1-2*jstep

       !...  extrapolate inside or outside

       x2=x_in(j2)
       x1=x_in(j1)
       x0=x_in(j0)
       xm=x_in(jm)

       aintm=((x-x1)*(x-x2)+(x-x0)*(x-x2)+(x-x0)*(x-x1)) &
            /((xm-x0)*(xm-x1)*(xm-x2))
       aint0=((x-x1)*(x-x2)+(x-xm)*(x-x2)+(x-xm)*(x-x1)) &
            /((x0-xm)*(x0-x1)*(x0-x2))
       aint1=((x-x0)*(x-x2)+(x-xm)*(x-x2)+(x-xm)*(x-x0)) &
            /((x1-xm)*(x1-x0)*(x1-x2))
       aint2=((x-x0)*(x-x1)+(x-xm)*(x-x1)+(x-xm)*(x-x0)) &
            /((x2-xm)*(x2-x0)*(x2-x1))

       dydx_out(j)=aintm*y_in(jm)+aint0*y_in(j0) &
            +aint1*y_in(j1)+aint2*y_in(j2)

    END DO

  end subroutine Lag3deriv_array


end module lagrange_interpolation
