import sys
import os
import math
import copy
import numpy as np
from ParIO import Parameters


class ProfileData(object):
    """Class handling the profile input, 
    averaging and assignment to numpy vectors for a single given extension
    special choices for starttime and endtime:
    -1: use first time in profile file
    -2: use last time in profile file
    """
    def __init__(self,fext, starttime, endtime, plotset):
        """ Read parameter file and create empty arrays for profile data """
        self.isDataPresent = False
        self.fext = fext
        self.plotset = copy.deepcopy(plotset) # each dataset has its own collection of desired plots
        self.locncpathnr = "000"    # directory extension for the local equilibrium data... not very nice yet
        self.starttime = starttime
        self.endtime = endtime
        self.numblocks = 0 #number of blocks read from the profile input file
        
        par=Parameters()
        par.Read_Pars('parameters%s' %(self.fext))
        self.nspec=int(par.pardict['n_spec'])
        #coll = float(par.pardict['coll'])
        self.rhostar = float(par.pardict['rhostar'])
        self.nz0 = int(par.pardict['nz0'])
        self.lbuffer = 0
        self.ubuffer = 0
        if ('l_buffer_size' in par.pardict):
            self.lbuffer = float(par.pardict['l_buffer_size'])
        if ('u_buffer_size' in par.pardict):
            self.ubuffer = float(par.pardict['u_buffer_size'])
        self.major_R = float(par.pardict['major_R'])
        self.minor_r = float(par.pardict['minor_r'])
        q_coeffs= str(par.pardict['q_coeffs'])
        self.q_coeff_list = map(float,q_coeffs.split(","))
        self.coll = float(par.pardict['coll'])
        self.specname=[[],[]]
        for n in range(self.nspec):
            self.specname[n] = str(par.pardict['name%1d'%(n+1)])
            self.specname[n] = self.specname[n][1:-1]
        self.istep_prof = int(par.pardict['istep_prof'])
        if ('plotfsa' in self.plotset and 'istep_fsa_moments' in par.pardict):
            self.istep_fsamom = int(par.pardict['istep_fsa_moments'])
        self.comp_type = str(par.pardict['comp_type'])
        self.nx0=int(par.pardict['nx0'])
        
        self.units= {'nref':1.0, 'Tref':1.0, 'Bref':1.0, 'Lref':1.0, 'mref':1.0}
        for k in self.units:
            if k in par.pardict:
                self.units[k] = float(par.pardict[k])
            
        self.xs=np.empty((self.nx0))
        self.Afs=np.empty((self.nx0))
        self.Gammas=np.zeros((self.nx0,self.nspec))
        self.Gammaturb=np.zeros((self.nx0,self.nspec))
        self.Qs=np.zeros((self.nx0,self.nspec))
        self.Qturb = np.zeros((self.nx0,self.nspec))
        self.Qcharr = np.zeros((self.nx0,self.nspec))
        self.jbs=np.zeros((self.nx0,self.nspec))
        self.Pis=np.zeros((self.nx0,self.nspec))
        self.Piturb=np.zeros((self.nx0,self.nspec))
        self.Ts=np.zeros((self.nx0,self.nspec))
        self.ns=np.zeros((self.nx0,self.nspec))
        self.omts=np.zeros((self.nx0,self.nspec))
        self.omns=np.zeros((self.nx0,self.nspec))
        self.T0s=np.empty((self.nx0,self.nspec))
        self.n0s=np.empty((self.nx0,self.nspec))
        self.omt0s=np.empty((self.nx0,self.nspec))
        self.omn0s=np.empty((self.nx0,self.nspec))
        self.nustar=np.zeros((self.nx0,self.nspec))
    
    def q_fun(self,x):
        qval = 0.0
        for i in xrange(0,len(self.q_coeff_list)-1):
            qval += self.q_coeff_list[i]*x**i
        return qval

    def _chang_hinton(self):
        """ Calculate the Chang-Hinton estimate for the neoclassical heat flux """
        for n in range(1): #only main ion species
            epsilon = np.array(self.xs*self.minor_r/self.major_R)
            mspec =1 #since we only support one ion species so far
            # alpha = Zeff - 1
            alpha = 0
            Z=1
            #for alpha == 0, mustar == nustar
            mustar = self.nustar[:,n]
            #(np.array(map(self.q_fun,self.xs))/epsilon**1.5 * self.major_R * 8/3/math.sqrt(math.pi) 
            #          * self.ns[:,n] / self.Ts[:,n]**2 * self.coll)
            f1 = 1 + 1.5*epsilon**2
            f2 = (1-epsilon**2)**0.5
            #k1 = 0 # relevant only for Zeff != 1
            k2 = (0.66 + 1.88*epsilon**0.5 - 1.54*epsilon)/(1+1.03* mustar**0.5 + 0.31* mustar)*f1 + \
                  1.6*0.74* mustar*epsilon/(1+0.74*mustar*epsilon**1.5)*0.5*(f1-f2)
            qchval = 16/3/math.sqrt(math.pi) * self.ns[:,n]**2 *\
                     np.array(map(self.q_fun,self.xs))**2/epsilon**1.5*(1-epsilon**2)*\
                     Z**4 * self.coll * np.sqrt(2*self.Ts[:,n]*mspec) * k2 * self.omts[:,n]
            self.Qcharr[:,n] = qchval

    def readprofiles(self):
        """ Read density, temperature, gradients and fluxes """
        for n in range(self.nspec):
            print('READ   profile_%s%s\n'%(self.specname[n],self.fext))
            proffile=open('profile_%s%s'%(self.specname[n],self.fext))
            lines=proffile.readlines()
            first_time=float(lines[1].split()[1])
            last_time=float(lines[len(lines) - 3 - self.nx0].split()[1])
            print self.endtime
            print self.starttime
            if (self.starttime == -1) or (self.starttime > 0 and self.starttime < first_time):
                print "Using first time present in profile data for starttime"
                self.starttime = first_time
            if (self.endtime == -1):
                print "Using first time present in profile data for endtime"
                self.endtime = first_time
            if self.starttime == -2:
                print "Using last time present in profile data for starttime"
                self.starttime = last_time
            if (self.endtime == -2) or (self.endtime > last_time):
                print "Using last time present in profile data for endtime"
                self.endtime = last_time
            if (self.endtime < first_time) or (self.starttime > last_time):
                print "Time window not contained in profile data"
                return
            lastblock = (len(lines) - 1)/(self.nx0 + 3)
            print 'starttime=%f, endtime=%f, first_time=%f, last_time=%f'%(self.starttime,self.endtime,first_time, last_time)
            self.numblocks = 0
            #1 header + (header+nx0 values)
            for block in xrange(1,lastblock):
                lstart = 2+(1+self.nx0+2)*block
                lstart2 = 2+(1+self.nx0+2)*(block-1)
                l = lstart - 1
                l2 = lstart2 - 1
                if float(lines[l].split()[1]) == float(lines[l2].split()[1]):
                    print "Error: Profile %s contains 2 blocks with identical timestamp"%(self.fext)
                    return
                read_time = float(lines[l].split()[1])
                prev_time = float(lines[l2].split()[1])
                timestep = (read_time - prev_time)
                if (self.starttime == self.endtime):
                    if (read_time >= self.starttime):
                        timestep = 1
                        timetemp = read_time
                        lstart2 = lstart
                        self.starttime = read_time
                        self.endtime = read_time
                        self._addprofdata(lines,n,lstart,lstart2,timestep)
                        break
                # do the average by using the trapezoid rule
                if (read_time >= self.starttime-timestep/4) and (read_time <= self.endtime+timestep/4):
                    self.numblocks += 1
                    timetemp = read_time
                    if self.numblocks == 1:
                        self.starttime = prev_time
                    self._addprofdata(lines,n,lstart,lstart2,timestep)
            self.endtime = timetemp
            print 'numblocks', self.numblocks
            prof0file=open('profiles_%s%s'%(self.specname[n],self.fext))
            lines = prof0file.readlines()
            for i in xrange(0,self.nx0):
                l=2+i
                self.xs[i]=lines[l].split()[0]
                self.T0s[i,n]=float(lines[l].split()[2])/self.units['Tref']
                self.n0s[i,n]=float(lines[l].split()[3])/self.units['nref']
                self.omt0s[i,n]=lines[l].split()[4]
                self.omn0s[i,n]=lines[l].split()[5]
            proffile.close()
            prof0file.close()
        totaltime = (self.endtime - self.starttime)
        if abs(totaltime) < 1e-8: #we only have one block
            totaltime = 1
        self.Gammas /= totaltime
        self.Gammaturb /= totaltime
        self.Qs /= totaltime
        self.Qturb /= totaltime
        self.Pis /= totaltime
        self.Piturb /= totaltime
        self.jbs /= totaltime
        self.Ts /= totaltime
        self.ns /= totaltime
        self.omts /= totaltime
        self.omns /= totaltime
        epsilon = np.array( [x/self.major_R*self.minor_r for x in self.xs])
        self.Qs -= 2.5* self.Gammas*self.Ts     #subtract convective heat flux
        self.Qturb -= 2.5*self.Gammaturb*self.Ts
        for n in range(self.nspec):
            self.nustar[:,n] = (8.0/3.0/math.sqrt(math.pi)*self.major_R *
                                self.coll / epsilon**(3.0/2)*map(self.q_fun,self.xs)*
                                self.ns[:,n]/self.Ts[:,n]**2)
        #flux surface area
        self.Afs=4*np.pi**2*self.major_R*self.minor_r*self.xs 
        #index of x0 position (x0 = 0.5)!
        self.x0ind = (np.abs(self.xs-0.5)).argmin()
        self._chang_hinton()
        self._readlocalfile()
        self.fsastarttime = self.starttime
        self.fsaendtime = self.endtime
        self._readfsadata()
        self.isDataPresent = True
        print 'Actual limits: starttime=%f, endtime=%f'%(self.starttime,self.endtime)

    def _addprofdata(self,lines,n, lstart,lstart2,timestep):
        for i in xrange(0,self.nx0):
            l=lstart+i
            l2=lstart2+i
            self.Gammas[i,n]+=0.5*(float(lines[l].split()[9])+float(lines[l2].split()[9]))*timestep
            self.Gammaturb[i,n]+=0.5*(float(lines[l].split()[6])+float(lines[l2].split()[6]))*timestep
            self.Qs[i,n]+=0.5*(float(lines[l].split()[10])+float(lines[l2].split()[10]))*timestep
            self.Qturb[i,n]+=0.5*(float(lines[l].split()[7])+float(lines[l2].split()[7]))*timestep
            self.Pis[i,n] +=0.5*(float(lines[l].split()[11])+float(lines[l2].split()[11]))*timestep
            self.Piturb[i,n] +=0.5*(float(lines[l].split()[8])+float(lines[l2].split()[8]))*timestep
            self.jbs[i,n]+=0.5*(float(lines[l].split()[12])+float(lines[l2].split()[12]))*timestep
            self.Ts[i,n]+=0.5*(float(lines[l].split()[2])+float(lines[l2].split()[2]))*timestep
            self.ns[i,n]+=0.5*(float(lines[l].split()[3])+float(lines[l2].split()[3]))*timestep
            self.omts[i,n]+=0.5*(float(lines[l].split()[4])+float(lines[l2].split()[4]))*timestep
            self.omns[i,n]+=0.5*(float(lines[l].split()[5])+float(lines[l2].split()[5]))*timestep


    def _readlocalfile(self):
        """Read the data from local runs set up with glob2loc"""
        locncfilename="glob2locscan%s/neo.log"%(self.locncpathnr)
        if (not os.path.isfile(locncfilename)):
            self.plotset.discard('plotlocQ')
            print "No local NC equilibrium present"
            print locncfilename
            return
        if ('plotlocQ' in self.plotset):
            self.locxpos=[]
            self.locqval=[]
            self.locjbval=[]
            #self.locchival=[]
            locncfile = open(locncfilename)
            lines=locncfile.readlines()
            for lin in lines:
                li = lin.strip()
                if (not li.startswith('#')):
                    self.locxpos.append(float(lin.split()[6]))
                    tmpnum = (float(lin.split()[9]))*(float(lin.split()[2]))*(float(lin.split()[4]))**2.5
                    tmpnum = tmpnum - 2.5*(float(lin.split()[8]))*(float(lin.split()[2]))*(float(lin.split()[4]))**2.5
                    self.locqval.append(tmpnum)
                    tmpnum = (float(lin.split()[11]))*(float(lin.split()[2]))*(float(lin.split()[4]))
                    self.locjbval.append(tmpnum)
            locncfile.close()

    def _readfsadata(self):
        """Read the data from the fsa_moments diagnostic"""
        fsamomfilename='fsamom_%s%s'%(self.specname[0],self.fext)
        if(not os.path.isfile(fsamomfilename)):
            self.plotset.discard('plotfsa')
            print "No fsamom file present!"
            print fsamomfilename
            return
        print "fsa moments diagnostic"
        if 'plotfsa' in self.plotset:
            nvars = 13
            self.fsamom = np.zeros((self.nx0,nvars))
            fsamomfile=open(fsamomfilename)
            lines=fsamomfile.readlines()
            lastblock = (len(lines) - 1)/(self.nx0 + 3)
            #1 header + (header+nx0 values)
            self.numblocksfsa = 0
            timetemp = self.fsaendtime
            for block in xrange(1,lastblock):
                lstart = 2+(1+self.nx0+2)*block
                lstart2 = 2+(1+self.nx0+2)*(block-1)
                l = lstart - 1
                l2 = lstart2 - 1
                if float(lines[l].split()[1]) == float(lines[l2].split()[1]):
                    print "Error: fsa_moments %s contains 2 blocks with identical timestamp"%(self.fext)
                    return
                read_time = float(lines[l].split()[1])
                prev_time = float(lines[l2].split()[1])
                timestep = (read_time - prev_time)
                # do the average by using the trapezoid rule
                if (read_time >= self.fsastarttime-timestep*0.9) and (read_time <= self.fsaendtime+timestep*0.9):
                    self.numblocksfsa += 1
                    if (self.fsastarttime == self.fsaendtime):
                        timestep = 1
                        lstart2 = lstart
                        timetemp = read_time
                        self.fsastarttime = read_time
                        self.fsaendtime = read_time
                    elif self.numblocksfsa == 1:
                        self.fsastarttime = prev_time
                    for i in xrange(self.nx0):
                        l=lstart+i
                        l2=lstart2+i
                        for n in range(nvars):
                            self.fsamom[i,n] += 0.5 *(float(lines[l].split()[n])+float(lines[l2].split()[n]))*timestep
                    timetemp = read_time
            self.fsaendtime = timetemp
            fsamomfile.close()
            print "numblocksfsa:", self.numblocksfsa
            print 'Actual limits (fsa): starttime=%f, endtime=%f'%(self.fsastarttime,self.fsaendtime)
            totaltime = self.fsaendtime - self.fsastarttime
            if abs(totaltime) < 1e-8: #we only have one block
                totaltime = 1
            self.fsamom /= totaltime
            dx=self.xs[1]-self.xs[0]
            self.intfsamom=np.zeros((self.nx0,nvars))
            print "Integrating"
            for n in range(nvars):
                for i in xrange(0,self.nx0-1):
                    self.intfsamom[i+1,n]=self.intfsamom[i,n]+self.Afs[i]*self.fsamom[i+1,n]*dx/self.rhostar
                    #intfsamom[i+1,n]=intfsamom[i,n]+fsamom[i+1,n]*dx/rhostar
                self.intfsamom[:,n]/=self.Afs

    def gluecontindata(self, contin):
        """ Add an additional set of data which starts right after this object """
        delta = 1e-6
        if not (self.nspec == contin.nspec and
           abs(self.rhostar - contin.rhostar) < delta and
           self.nz0 == contin.nz0 and
           abs(self.major_R - contin.major_R) < delta and
           abs(self.minor_r - contin.minor_r) < delta and
           #self.q_coeff_list =     ### TODO: check q profile
           abs(self.coll - contin.coll) < delta and
           self.specname == contin.specname and
           #self.istep_prof == contin.istep_prof and
           self.comp_type == contin.comp_type and
           self.nx0 == contin.nx0):
            print "Parameter mismatch"
            return False
        if (self.endtime - contin.starttime) > 0:
            self.endtime = contin.starttime
            self.readprofiles()
            print "Times overlap. Adjusting... (profile%s takes precedence)"%(contin.fext)
        if abs(self.endtime - contin.starttime) > delta:
            print "Times of the profiles do not match!"
            return False

        firsttotal = self.endtime - self.starttime
        secondtotal = contin.endtime - contin.starttime
        totaltime = firsttotal + secondtotal
        
        
        self.Gammas = (self.Gammas*firsttotal + contin.Gammas*secondtotal)/totaltime
        self.Gammaturb = (self.Gammaturb*firsttotal + contin.Gammaturb*secondtotal)/totaltime 
        self.Qs = (self.Qs*firsttotal + contin.Qs*secondtotal)/totaltime
        self.Qturb = (self.Qturb*firsttotal + contin.Qturb*secondtotal)/totaltime
        self.Pis = (self.Pis*firsttotal + contin.Pis*secondtotalt)/totaltime
        self.Piturb = (self.Piturb*firsttotal + contin.Piturb*secondtotalt)/totaltime
        self.jbs = (self.jbs*firsttotal + contin.jbs*secondtotal)/totaltime
        self.Ts = (self.Ts*firsttotal + contin.Ts*secondtotal)/totaltime
        self.ns = (self.ns*firsttotal + contin.ns*secondtotal)/totaltime
        self.omts = (self.omts*firsttotal + contin.omts*secondtotal)/totaltime
        self.omns = (self.omns*firsttotal + contin.omns*secondtotal)/totaltime
        epsilon = np.array( [x/self.major_R*self.minor_r for x in self.xs])
        for n in range(0, self.nspec-1):
            self.nustar[:,n] = (8.0*math.sqrt(2)/3.0/math.sqrt(math.pi)*self.major_R *
                            self.coll / epsilon**(0.0/2)*map(self.q_fun,self.xs)*
                            self.ns[:,n]/self.Ts[:,n]**2)
        self._chang_hinton()
        
        if ('plotfsa' in self.plotset):
            if ('plotfsa' in contin.plotset):
                firsttotal = self.fsaendtime - self.fsastarttime
                secondtotal = contin.fsaendtime - contin.fsaendtime
                totaltime = firsttotal + secondtotal
                self.fsamom = (self.fsamom*firsttotal + contin.fsamom*secondtotal)/totaltime
                self.intfsamom = (self.intfsamom*firsttotal + contin.intfsamom*secondtotal)/totaltime
                self.fsaendtime = contin.fsaendtime
            else:
                print "Removing fsa moments"
                self.plotset.discard('plotfsa')
        self.endtime = contin.endtime
        print "Profile combination completed"
        return True

    def avprof_output(self,avproffile_prefix):
        print "Writing averaged profile to file"
        for n in range(0,self.nspec):
            avprofilename = (avproffile_prefix + "_" + self.specname[n])#.strip()
            avproffile=open(avprofilename, 'w')
            avproffile.write('#   x/a             x/a(dummy)       T/Tref          n/nref\n')
            avproffile.write('# Averaged profile generated by ProfileData.py: t = %4.1f - %4.1f\n' %(self.starttime,self.endtime))
            for i in xrange(0,self.nx0):
                avproffile.write(str('%E'%self.xs[i]) + "\t" + str('%E'%self.xs[i]) + "\t")
                avproffile.write(str('%E'%self.Ts[i,n]) + "\t" + str('%E'%self.ns[i,n]) + "\n")
            avproffile.close()
